; Accessors.
;
; Representation: Each <record> and <type> is represented exactly like it 
; was in the input _except_ for the first element.  The element is either 
; the symbol like it was in the input, or a pair.  If it is a pair, then 
; the car of the pair is the symbol from the input and the cdr is 
; system-internal information.  It's a list which currently holds:
;    (referenced-bit cached-info)
;
; All other data are represented exactly as in the input.
;
; Some of the generic functions operate on data structures which do not
; have a record-tag (fields, for example).  They examine the datum to figure
; out what to do.  For example, if the car is a string then it's a field,
; otherwise it's not.

;; Record uniform/generic accessors.  These are accessors which should
;; work on the <record> items.


;; Always use this to get the tag from a record.

;; The records are modified in-place to add some processing
;; information.  This routine will safely access the tag, regardless
;; of whether the record has been modified or not.

(define (record-tag rec)
  (if (symbol? (car rec))
      (car rec)
      (caar rec)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; general <record> accessors.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; file name in records which have one

(define (get-filename file-info)
  (car (list-ref file-info 1)))

;; get the source file line number

(define (get-line-number file-info)
  (cadr (list-ref file-info 1)))

;; get the name associated with the record

(define (get-name rec)
  (list-ref rec 2))

;; type in records which have one

(define (get-type rec)
  (list-ref rec 3))

;; attrs in records which have one

(define (get-attrs rec)
  (list-ref rec 4))

;; for macro records

(define (get-macro-name rec)
  (list-ref rec 2))

(define (get-macro-def rec)
  (list-ref rec 3))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; structure and union accessors
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; for struct/union/enum/typedef references

(define get-ref-tag cadr)

;; for structure/union fields

(define (get-struct-fields rec)
  (list-ref rec 3))

;; is it a field or a bit-field ?

(define get-field-tag caadr)

(define get-field-name car)

(define get-field-type cadadr)

(define (get-field-size field-info)
  (list-ref field-info 2))

(define (get-field-offset field-info)
  (list-ref field-info 3))

(define (get-bit-field-info rec)
  (list-ref rec 4))

;; For bit-field specific information the bit-field specific info
;; should be used, i.e. the value returned by applying bit-field-info
;; to a record.

(define get-bit-field-loc car)

(define get-bit-field-start cadr)

(define (get-bit-field-end x)
  (list-ref x 2))

(define (get-bit-field-mask x)
  (list-ref x 3))
  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; enums
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (get-enum-info rec)
  (list-ref rec 3))

(define get-enum-name car)

(define get-enum-value cadr)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; function accessors
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; function argument list (list  of types)

(define get-function-arglist cadr)

;; function return type

(define get-function-return-type caddr)

;; pointer types information

(define get-pointer-type cadr)

;; array info

(define get-array-length cadr)

(define get-array-type caddr)

;; type operations

(define get-type-tag car)

;; only valid when used on a primitive type !

(define get-type-attrs cadr)

;; The following accessors are used for internal data structures.

(define (sysinfo rec)
  (if (symbol? (car rec))
      (let ((info (list #f '())))
	(set-car! rec (cons (car rec) info))
	info)
      (cdar rec)))

;; Given a type record, strip all of the attributes from it so that
;; only the type information is left.

(define strip-attrs
  (lambda (type)
    (let ((tt (type-tag type)))
      (cond
       ((primitive-type? tt)
	(list (car type)))
       ((eq? 'pointer tt)
	(list 'pointer (strip-attrs (pointer-type type))))
       (else
	type)))))

;; $Id: accessors.scm,v 1.2 2001/02/02 06:51:02 briand Exp $
