;; Supports routines which process macros from an ffi file.

;; to determine if a macro-name is simple:
;;
;; scan the characters and if no whitespace is found, the macro is
;; simple.


(define simple?
  (lambda (macro-name)
    (let loop ((i 0))
      (cond
       ((>= i (string-length macro-name))
	#t)
       ((char-whitespace? (string-ref macro-name i))
	#f)
       (else
	(loop (+ i 1)))))))

;; Notice that there is no name-mangling.  It's probably a good idea
;; to leave these upper case so that the user of the ffi knows they
;; are upper case.  Personally I think that scheme SHOULD be
;; case-sensitive.
;;
;; There are a lot of heurisitcs which go into selecting the macros to
;; process :
;;
;; macros which have the same name and definition are ignored (I don't
;; know why these show up), e.g.
;;
;; #define SOCK_DGRAM SOCK_DGRAM
;;
;; Macros names which begin with __ are ignored.
;;
;; macros with an empty definition are defined as true to indicate
;; that they are defined.
;;
;; macros that begin with and end with " are strings and returned as
;; such.
;;
;; Only "simple" macros are processed.  Macros which have arguments are
;; returned as strings.

(define macro-string?
  (lambda (def)
    (cond
     ((< (string-length def) 2)
      #f)
     ((and (char=? #\" (string-ref def 0))
	   (char=? #\" (string-ref def (- (string-length def) 1))))
      def)
     (else
      #f))))

(define process-macros
  (lambda (macros)
    (map
     (lambda (macro)
       (let* ((mname (get-macro-name macro))
	      (mdef (get-macro-def macro))
	      (mdef-type
	       (cond
		((string=? "" mdef)
		 #t)
		((string->number mdef)
		 (string->number mdef))
		((macro-string? mdef)
		 mdef)
		(else
		 mdef))))
	 `(define ,(string->symbol mname) ,mdef-type)))
     (filter
      (lambda (macro)
	(let ((name (get-macro-name macro))
	      (def (get-macro-def macro)))
	  (and (not (prepends__? name))
	       (not (string=? name def))
	       (simple? name))))
      macros))))

;; $Id: ffi-macros.scm,v 1.4 2001/02/07 07:50:21 briand Exp $
