#! /usr/local/bin/gsi -f

(load "../srfi-1.scm")
(load "../ffigen.scm")

(define err-exit
  (lambda args
    (for-each
     (lambda (arg)
       (display arg))
     args)
    (exit 1)))

(define ref-table #f)

(define primitive-types 
  '((void . "void")
    (char . "char") 
    (unsigned-char . "unsigned char")
    (signed-char . "signed char ")
    (short . "short")
    (unsigned-short . "unsigned short")
    (unsigned . "unsigned")
    (int . "int")
    (unsigned-int . "unsigned int ")
    (long . "long")
    (unsigned-long . "unsigned long")
    (long-long . "long long")
    (unsigned-long-long . "unsigned long long")
    (float . "float")
    (double . "double")
    (long-double . "long double))")))


(define primitive-type?
  (lambda (type)
    (assq type primitive-types)))

(define c-type
  (lambda (type)
    (cdr (assq type primitive-types))))

;; Given a type from the intermediate representation file, return a
;; gambit-ffi equivalent type.
;;
;; when deref-typedef? is true, take typedefs and reduce them to their
;; equivalent type's, i.e. struct's, pointers, etc., otherwise simply
;; return the typedef name.
;;
;; typedef's are quite the problem.  In some instances, like function
;; arg declarations, we want to use the typedef name, so that the
;; gambit ffi will create c-declarations using the typedef name.
;;
;; However, when creating the c-define types the typdef must be
;; "reduced" since the typedef used in the function declarations must
;; reducible sofgambit knows what type of transformation to apply to
;; convert to and from scheme values.
;;
;; This routine was quite the headache to get right.

(define gambit-type 
  (letrec
      ((lookup-ref
	(lambda (type deref-typedef?)
	  (let ((t-tag (get-type-tag type))
		(ref-tag (get-ref-tag type)))
	    (if (memq t-tag '(struct-ref union-ref))
		(let* ((index (string->number ref-tag))
		       (designator
			(if index
			    (let ((name (cdr (vector-ref ref-table index))))
			      (if deref-typedef?
				  name
				  (string->symbol name)))
			    (get-ref-tag type))))
		  (case t-tag
		    ((struct-ref)
		     `(struct ,designator))
		    ((union-ref)
		     `(union ,designator))
		    (else
		     (err-exit "lookup-ref: unknown type tag " t-tag))))
		(err-exit "lookup-ref: " 
			  t-tag " is not a struct/union-ref")))))
       (get-type
	 (lambda (type . deref-typedef?)
	   (let ((type-tag (get-type-tag type)))
	     (cond
	      ((eq? 'unsigned type-tag)
	       'unsigned-int)
	      ((primitive-type? type-tag)
	       type-tag)
	      (else
	       (case type-tag
		 ((typedef)
		  (if (null? deref-typedef?)
		      (string->symbol (get-ref-tag type))
		      (get-type (reduce-typedef type types))))
		 ((struct-ref union-ref)
		  (lookup-ref type deref-typedef?))
		 ((enum-ref)
		  'int)
		 ((pointer)
		  `(pointer ,(apply get-type (get-pointer-type type) 
				    deref-typedef?)))
		 ((function)
		  (let ((arg-types (get-function-arglist type))) 
		    `(function
		      (,(apply get-type (car arg-types) deref-typedef?)
		       ,@(map 
			  (lambda (type)
			    (apply get-type type deref-typedef?))
			  (cdr arg-types)))
		      ,(get-type (get-function-return-type type)))))
		 ((array)
		  `(pointer ,(get-type (get-array-type type))))
		 (else
		  (err-exit 
		   "gambit-type: Cannot translate type: " type "\n")))))))))
    (lambda (type . deref-typedef?)
      (let ((result (apply get-type type deref-typedef?)))
	(cond
	 ;; (pointer char) is a special gambit-type : char-string
	 ((and (pair? result)
	       (eq? 'pointer (car result)))
	  (cond 
	   ((eq? 'char (cadr result))
	    'char-string)
	   ;; Gambit handles function types as pointers implicitly, so
	   ;; we need to strip the first "pointer" away from (pointer
	   ;; function) types.  IOW, (pointer (function ...)) should
	   ;; be returned to Gambit as simply (function ...)
	   ((and (pair? (cadr result))
		 (eq? 'function (caadr result)))
	    (cadr result))
	   (else
	    result)))
	 (else
	  result))))))

(define to-string
  (lambda (obj)
    (with-output-to-string
      (lambda ()
	(if (pair? obj)
	    (for-each
	     (lambda (item)
	       (display-line item " "))
	     obj)
	    (display obj))))))

;; Return a string which is the equivalent C-type of TYPE.  Useful for
;; generating ffi code directly in C.
;;
;; Type is an ffi type and not "gambit-type".

(define type->c-type
  (lambda (type)
    (let ((type-tag (get-type-tag type))
	  (ref-tag (get-ref-tag type)))
      (cond
       ((eq? 'unsigned type-tag)
	"unsigned int")
       ((primitive-type? type-tag)
	(c-type type-tag))
       (else
	(case type-tag
	  ((typedef)
	   ref-tag)
	  ((struct-ref)
	   (to-string (lookup-ref type)))
	  ((union-ref)
	   (to-string (lookup-ref type)))
	  ((enum-ref)
	   (to-string (lookup-ref type)))
	  ((pointer)
	   ;; pointers treated specially.  (pointer char) is rewritten
	   ;; to char-string which is a built-in gambit type.
	   (let ((p-type (cadr type)))
	     (string-append (type->c-type p-type) "*")))
;	  ((function)
;	   (let ((arg-list (cadr type))
;		 (ret-type (caddr type)))
;	     `(,(type->c-type ret-type) (*)(,(map type->c-type arg-list)))))
	  (else
	   (err-exit "type->c-type : Cannot translate type: " type "\n")
	   (exit))))))))

;; A function which, given a typdef type, will "follow" the typedef
;; until it has been reduced to a composition of primitive c types.
;; This is slightly harder than it seems since the typedef may be a
;; composition of typedefs.
;;
;; FIXME: maybe ? I don't _think_ circular typedefs are possible ?
;;
;; I need to cache the references to the typedefs, since they could be
;; easily called multiple times....

(define reduce-typedef
  (lambda (type types)
    (letrec ((reduce
	      (lambda (type)
		;; look for an occurence of the typedef in the typedef
		;; records.
		(if (primitive-type? (get-type-tag type))
		    type
		    (case (get-type-tag type)
		      ((typedef)
		       ;; Look-up failed - Something is seriously wrong.
		       (let ((def (lookup (get-ref-tag type) types)))
			 (if (not def)
			     (begin
			       (display-line "ERROR: " type " referenced.\n")
			       (display "but not found in type records.\n")
			       (exit 1))
			     (reduce (get-type def)))))
		      ((pointer)
		       `(pointer ,(reduce (get-pointer-type type))))
		      ((array)
		       `(array ,(reduce (get-array-type type))))
		      ((function)
		       (list 'function
			     (map reduce (get-function-arglist type))
			     (reduce (get-function-return-type type))))
		      ((enum-ref union-ref struct-ref)
		       type)
		      (else
		       ;; struct-ref union-ref. Substitute
		       ;; the reference.
		       (err-exit "reduce-typedef: Unhandled type " type)))))))
      (reduce type))))
		  
;; Check TYPE to see if it is on of the types listed in TYPES-LS.

(define of-type?
  (lambda (type types-ls)
    (let ((type (get-type-tag (if (eq? 'typedef (get-type-tag type))
				  (reduce-typedef type types)
				  type))))
      (let ((result (memq type types-ls)))
	result))))

;; Find all of the types which refer to a "number" tag.  Keep track of
;; the largest value reference.  When done create a vector of the
;; references with the corresponding typedef name, e.g,
;;
;; #((ref-number name) ...)
;;
;; So something like 
;;
;; (type ("/usr/include/bits/types.h" 132) 
;;  "__fsfilcnt64_t" 
;;  (struct-ref "4"))
;;
;; (type ("/usr/include/fftw.h" 54) 
;;  "fftw_complex" 
;;  (struct-ref "7"))
;;

(define make-ref-table
  (lambda (types)
    ;; Find all of the "anonymous reference types, i.e. references
    ;; that use a number as a label.  Anything which has a name will
    ;; be a typedef reference.
    (let ((ref-table
	   (let loop ((types types)
		      (ls '()))
	     (if (null? types)
		 (reverse ls)
		 (let* ((type-record (car types))
			(name (get-name type-record))
			(c-type (get-type type-record))
			(tag (get-type-tag c-type)))
		   (if (and (memq tag '(struct-ref enum-ref union-ref))
			    (string->number (get-ref-tag c-type)))
		       (loop (cdr types) (cons (cons c-type name) ls))
		       (loop (cdr types) ls)))))))
      ;; find the largest number index
      (let* ((max-index
	      (do ((ls ref-table (cdr ls))
		   (bign 0 (let ((n (string->number (get-ref-tag (caar ls)))))
			     (if (> n bign)
				 n
				 bign))))
		  ((null? ls) bign)))
	     (vec (make-vector (+ max-index 1))))
	(do ((i 0 (+ i 1))
	     (ls ref-table (cdr ls)))
	    ((null? ls)
	     vec)
	  (let ((item (caar ls)))
	    (vector-set! vec (string->number (get-ref-tag item)) 
			 (car ls))))))))

;; Lookup a type reference.  Either the reference has a name or it is
;; "anonymous", i.e. represented by a number.
;;
;; If it has a name, e.g. (struct-ref "struct-name") then the type
;; simply reduces to (struct "name").
;;
;; If it is something like (struct "2"), the the structure is an
;; anonymous definition, and is referred to by a typedef name.  In
;; this case, the typedef name is returned.
;;
;; Returns the appropriate gambit-type equivalent.  So a a named
;; structure will return :
;;
;; (struct "name")
;;
;; and a typedef name, will return a symbol, e.g.
;;
;; 'comples

(define lookup-ref
  (lambda (type)
    (let ((t-tag (get-type-tag type)))
      (if (not (memq t-tag '(struct-ref union-ref)))
	  (err-exit "lookup-ref : Unhandled type tag " t-tag))
      (let* ((ref-tag (get-ref-tag type))
	     (index (string->number ref-tag)))
	(if index
	    (let ((entry (vector-ref ref-table index)))
	      (and entry (string->symbol (cdr entry))))
	    `(,(cdr (assq t-tag '((struct-ref . struct) (union-ref . union))))
	      ,ref-tag))))))
      
(define in-ref-table?
  (lambda (type)
    (and (memq (get-type-tag type) '(struct-ref union-ref))
	 (string->number (get-ref-tag type)))))

;; Create a list of Gambit style ffi's for functions.

;; This routine is mainly intended to create c-define types to
;; represent c typedefs.
;; 
;; Write out c-define-types for all of the types which 
;;
;; In particular, take the types which are in only-files and makes
;; sure that "supporting" type definitions are included.

(define make-type-ffi
  (lambda (types)
    (map
     (lambda (type-record)
       (let* ((name (get-name type-record))
	      (c-type (get-type type-record))
	      (tag (get-type-tag c-type))
	      (ref (get-ref-tag c-type)))
	 `(c-define-type ,(string->symbol name) ,(gambit-type c-type #t))))
     (filter referenced? types))))

;; functions : ffi function records
;;
;; This routine generates function names which start with %.  IME,
;; it's often useful to be able to shadow the names of the c function
;; with a scheme function, so the c-interface routines needs to sit in
;; a separate namespace.
;;
;; It's important for this routine to check the argument types and
;; return type since it is illegal to pass or return struct/union's.
;;
;; Right now only the return type is checked.  Even when it is not
;; checked, the gambit compiler catches the error.

(define make-function-ffi
  (lambda (functions)
    (map
     (lambda (function-record)
       (let* ((name (get-name function-record))
	      (scheme-name (string->symbol (c-name->scheme-name name "%")))
	      (function (get-type function-record))
	      (arg-list (get-function-arglist function))
	      (ret-type (get-function-return-type function)))
	 ;; check for valid return type also, a void in the arglist
	 ;; indicates a varargs condition, and must be trapped also.
	 (if (not (or (of-type? ret-type '(struct-ref union-ref))
		      (any 
		       (lambda (type) 
			 (of-type? type '(struct-ref union-ref void)))
		       arg-list)))
	     
	     ;; Assemble a c-lambda representation using the function
	     ;; info.
	     `(define ,scheme-name
		(c-lambda 
		 ,(map (lambda (arg) (gambit-type arg)) arg-list)
		 ,(gambit-type ret-type)
		 ,name))
	     (begin
	       (if (of-type? ret-type '(struct-ref union-ref void))
		   (warn
		    (string-append "Function '" 
				   name 
				   "' return type is 'struct' or 'union' ")
		    "which cannot be returned from a c-lambda.")
		   (warn
		    (string-append "Function '" 
				   name 
				   "' function argument is type 'void'")
		    "which indicates varargs call (can't be handled)."))
	       `(define ,scheme-name #f)))))
     ;; filter based on the function name.  Most scheme names which
     ;; begin with more than one '_' are meant to be "internal"
     ;; names.
     functions)))

;; Create ffi's used to access global variables by creating a wrapper
;; function which simply returns the value of the global variable.
;; Also, just to be useful, create a wrapper which can set the
;; variable too.

(define make-var-ffi
  (lambda (vars)
    (append-map
     (lambda (var-record)
       (let* ((name (get-name var-record))
	      (type (get-type var-record))
	      (c-name (c-name->scheme-name name)))
	 (if (not (of-type? type '(union-ref struct-ref)))
	     `((define ,(string->symbol (string-append c-name "-set!"))
		 (c-lambda 
		  (,(gambit-type type))
		  void
		  ,(string-append 
		    " extern " (type->c-type type) " " name ";"
		    " " name " = ___arg1;")))
	       (define ,(string->symbol c-name)
		 (c-lambda 
		  ()
		  ,(gambit-type type)
		  ,(string-append 
		    " extern " (type->c-type type) " " name ";"
		    " ___result = " name ";"))))
	     (begin
	       (warn
		(string-append 
		 "Variable '" name "' is type 'struct' or 'union' ")
		"which cannot be returned from a c-lambda.")
	       `((define ,(string->symbol c-name) #f))))))
     vars)))

;; This routine is responsible for creating the procedures which
;; manipulate c structs.  It generates both accessor code and mutator
;; code.  The names follow the convention for the gambit-c native
;; structures.
;;
;; Figuring out a name for a structure is a little tricky.  If the
;; structure has been declared "normally", i.e. as struct name, then
;; everything is straightforward.
;;
;; If the structure is referred to anonymously, i.e. only exists as a
;; typedef, then the declaration must use the typedef name, without a
;; "struct" prefix.
;;
;; Naturally the structure field must be checked to see if it is, in
;; fact, another structure.  If it is, or a union, then the accessor
;; functions can't be generated.
;;
;; FIXME: structures embedded in structures are not handled properly.
;; We just skip them for now.

;; Needed to derive a name for a structure/union.  Either the
;; structure or union already has name, easy...  or it is a reference
;; number.  If it's a reference # then de-reference it and use the
;; typedef name.

(define find-ref-name
  (lambda (ref-tag)
      (let* ((index (string->number ref-tag)))
	(if index
	    (let ((entry (vector-ref ref-table index)))
	      (and entry (cdr entry)))
	    ref-tag))))

(define make-struct-ffi
  (let* (

	 ;; And last but certainly not least, create a constructor for
	 ;; the structure, but create one which automagically GC's too
	 ;; !
	 (make-allocator
	  (lambda (type-name type scheme-type)
	    (let ((allocator-name '%new)
		  (c-type-name 
		   (type->c-type type))
		  (ptr-type-name 
		   (type->c-type `(pointer, type))))
	      `(define ,(string->symbol (string-append "%make-" type-name))
		 (let ((,allocator-name
			(c-lambda () (pointer ,scheme-type)
				  ,(string-append
				    " ___result = "
				    "(" ptr-type-name ")" 
				    "malloc (sizeof(" c-type-name "));"
				    "assert(___result);"))))
		   (lambda ()
		     (let* ((obj (,allocator-name)))
		       (make-will obj (lambda ()
					(display "FREE")
					(%free obj)))
		       obj)))))))
	 
	 (process-struct
	  (lambda (struct-rec name struct-name)
	    (let* ((fields (get-struct-fields struct-rec))
		   (ffi-type `(struct-ref ,name))
		   (struct-decl (lookup-ref ffi-type))
		   (process-field
		    (lambda (field field-type)
		      (let* ((field-name (get-field-name field))
			     (make-getter
			      ;; generate the field accessor - very
			      ;; straightforward.  The argument MUST be a
			      ;; pointer to the structure.
			      (lambda (getter-name)
				`(define ,getter-name
				   (c-lambda 
				    ((pointer ,struct-decl))
				    ,(gambit-type field-type)
				    ,(string-append 
				      " ___result = " 
				      "___arg1->" field-name ";")))))
			     ;; generate the field setter
			     (make-setter
			      (lambda (setter-name)
				`(define ,setter-name
				   (c-lambda 
				    ((pointer ,struct-decl) 
				     ,(gambit-type field-type))
				    void
				    ,(string-append " ___arg1->" field-name 
						    " = ___arg2;"))))))
			(let* ((c-name (c-name->scheme-name field-name))
			       (c-struct-name 
				(c-name->scheme-name struct-name))
			       (getter-name 
				(string->symbol
				 (string-append "%" struct-name "-" c-name)))
			       (setter-name 
				(string->symbol
				 (string-append "%" c-struct-name "-" c-name 
						"-set!"))))
			  (list
			   (make-getter getter-name)
			   (make-setter setter-name)))))))
	      (cons
	       (make-allocator struct-name ffi-type struct-decl)
	       (append-map 
		(lambda (field)
		  (let ((field-tag (get-field-tag field)))
		    (if (eq? 'field field-tag)
			(let ((field-type (get-field-type field)))
			  (if (of-type? field-type 
					'(struct-ref union-ref array))
			      (begin
				(warn
				 (string-append 
				  "Field '" (get-field-name field)
				  " in " struct-name "' is type ")
				 field-type
				 "which cannot be returned from a c-lambda.")
				'())
			      (process-field field field-type)))
			(begin
			  (display "WARNING: can't handle bit fields yet...\n")
			  '()))))
		fields))))))
    (lambda (structs skip-structs)
      (append-map 
       (lambda (struct-rec)
	 (let* ((name (get-name struct-rec))
		(struct-name (find-ref-name name)))
	   (if struct-name
	       (process-struct struct-rec name struct-name)
	       '())))
       (filter 
	(lambda (struct)
	  ;; don't process unreferenced structures, or structures
	  ;; without fields.
	  (and (referenced? struct)
	       (not (null? (get-struct-fields struct)))))
	structs)))))

;; Create enum ffi code using ENUMS.  Ignores enums which start with
;; "__".

(define make-enum-ffi
  (lambda (enums)
    (append-map
     (lambda (enum-record)
       (let* ((enum-info (get-enum-info enum-record)))
	 (map
	  (lambda (enum-def)
	    (let ((name (get-enum-name enum-def))
		  (value (get-enum-value enum-def)))
	      `(define ,(string->symbol (c-name->scheme-name name)) ,value)))
	  (filter 
	   (lambda (enum-def)
	     (let ((name (get-enum-name enum-def)))
	       (not (prepends__? name))))
	   enum-info))))
     enums)))

(define display-section
  (lambda (ls header  . newline-flag)
    (if (not (null? ls))
	(begin
	  (display-line "\n;;----- "header " -----\n\n")
	  (for-each
	   (lambda (item)
	     (pp item)
	     (newline))
	   ls)))))

;; process the skip file and initialize the "skip-lists".
;;
;; TODO: read in the list from skip.scm and establish the skip-lists
;; for the different types.


;; only-files : A list of filenames.  Process records which originate
;; in files in the only-files list, ignoring all others.
;;
;; func-names : a list of functions names (the original function name)
;; which we're interested in.  All other function names will be
;; ignored.

(define skip-structs
  '("_XGC"
    "_XDisplay"
    "_XrmHashBucketRec"
    "_XOM"
    "_XOC"
    "_XIM"
    "_XIC"
    "_XRegion"))

(define write-gambit-ffi
  (lambda (only-files . func-names)
    (let ((func-names
	   (if (null? func-names)
	       '()
	       (car func-names))))
      (let* ((macros (apply only-records macros only-files))
	     (enums (apply only-records enums only-files))
	     (functions (select-records functions only-files func-names))
	     (vars (select-records vars only-files '())))
	;; Compute the referenced types based on the functions and
	;; variables of interest.
	(compute-referenced-types functions types)
	(compute-referenced-types vars types)
	;; This shouldn't be necessary...
;	(compute-referenced-types types types)
	(let ((macros-ffi (process-macros macros))
	      (enum-ffi (make-enum-ffi enums))
	      (types-ffi (make-type-ffi types))
	      (function-ffi (make-function-ffi functions))
	      (var-ffi (make-var-ffi vars))
	      (structs-ffi (make-struct-ffi structs skip-structs)))
	  ;; generate c-declare which will include the header file.
	  (display "(c-declare\n\"\n")
	  (for-each
	   (lambda (header-file)
	     (display-line "#include \\\"" header-file "\\\"\n"))
	   only-files)
	  (display "#include <assert.h>")
	  (display-line "\")\n")

	  (display '(define %free (c-lambda ((pointer void)) void "\"free\"")))

	  (display-section macros-ffi "C MACROS")
	  (display-section enum-ffi "ENUMS")
	  (display-section types-ffi "TYPE DEFINITIONS")
	  (display-section var-ffi "GLOBAL VARIABLES")
	  (display-section structs-ffi "STRUCTURE ACCESSORS")
	  (display-section function-ffi "FUNCTIONS"))))))

;; $Id: gambit-ffi.scm,v 1.5 2001/02/09 07:16:09 briand Exp $
